/*
 * Keypad driver for HA-N1
 *
 * Copyright 2014 Buffalo Inc.
 *
 * Licensed under the Proprietary.
 */

#include <linux/platform_device.h>
#include <linux/input-polldev.h>
#include "input_device_server.h"

#define POLL_INTERVAL_GPIO 20
#define POLL_INTERVAL_MIN 100
#define POLL_INTERVAL_MAX 1000

static const struct input_id info = {
		.bustype = BUS_VIRTUAL,
		.vendor  = 0x1,//!TODO
		.product = 0x1,//!TODO
		.version = 100,
};

//key_down_listの定義と同じ順に
const int key_code_array[key_down_all_list_count]={PANEL_LEFT,PANEL_RIGHT,PANEL_UP,PANEL_DOWN,PANEL_POWER_OFF,PANEL_HOME,UI_PROGRESS,UI_IMPORT_1,UI_IMPORT_2,UI_EXPANSION,UI_DMR,UI_NOTIFY,UI_USBODD_1,UI_USBODD_2};

/**
 * 指定のキーが実際に押されたままかチェック
 * キューではない
 */
bool is_key_down(volatile union key_down_uni *key,u16 key_code){
	int i;
	for(i=0;i<key_down_poll_list_count;i++){
		if(key_code_array[i]==key_code){
			return key->array.val[i];
		}
	}
	return false;
}

/**
 * キューのキーを取得
 */
u16 get_key(volatile union key_down_uni *key){
	int i;
	for(i=0;i<key_down_poll_list_count;i++){
		if(key->array.val[i]){
			return key_code_array[i];
		}
	}
	return SYN_REPORT;
}

/**
 * キューにキーがある
 */
bool has_active_key(volatile union key_down_uni *key){
	int i;
	for(i=0;i<key_down_poll_list_count;i++){
		if(key->array.val[i]){
			return true;
		}
	}
	return false;
}

/**
 * すべてのキーをクリア
 */
void clear_all_key(volatile union key_down_uni *key){
	memset((void*)key->array.val,false,sizeof(bool)*key_down_poll_list_count);
}

/**
 * ONになったものがあった場合に限り、キーをキューに追加
 */
void set_key_override(volatile union key_down_uni *from,volatile union key_down_uni *to){
	int i=0;
	for(i=0;i<key_down_poll_list_count;i++){
		if(from->array.val[i]){
			to->array.val[i]=true;
		}
	}
}

/**
 * ポーリング
 */
static void _poll(struct input_polled_dev *pdev){
	struct input_device_server_dev *bdev = pdev->private;
	struct input_dev *idev = bdev->input;
	volatile union key_down_uni *key_buffer = bdev->key_buffer;
	volatile union key_down_uni key_current={0};
	int code,value;
	unsigned int type = EV_KEY;

	spin_lock(&bdev->lock);

	pdev->poll_interval = idev->rep[REP_PERIOD];
#ifdef DEBUG
	printk( KERN_INFO "%s : _poll\n" ,DEV_MSG);
#endif
	if(0<idev->repeat_key){
		read_pin_values(&key_current);
		if(is_key_down(&key_current,idev->repeat_key)){
			//キーが押されている状態が継続している
			if(time_after(jiffies,bdev->last_keydown+(idev->rep[REP_DELAY] * HZ / 1000))){
				//最初のリピートはdelayさせる
				code  = idev->repeat_key;
				value = 2;//repert
				input_event(idev,type,code,value);
				input_sync(idev);
				if(key_buffer->list.key_poweroff){
					//powerキーに限り、リピート加速を行わない
					pdev->poll_interval = idev->rep[REP_DELAY];
				}
#ifdef DEBUG
				printk( KERN_ERR "%s : key repert %d\n" ,DEV_MSG,code);
#endif
			}
		} else {
			code  = idev->repeat_key;
			value = 0;//key_up
			input_event(idev,type,code,value);
			input_sync(idev);
			idev->repeat_key = 0;
			clear_all_key(key_buffer);
#ifdef DEBUG
			printk( KERN_ERR "%s : key up %d\n" ,DEV_MSG,code);
#endif
		}
	} else {
		if(has_active_key(key_buffer)){
			code  = get_key(key_buffer);
			value = 1;//key_down
			input_event(idev,type,code,value);
			input_sync(idev);
			idev->repeat_key = code;
			bdev->last_keydown = jiffies;
#ifdef DEBUG
			printk( KERN_ERR "%s : key down %d\n" ,DEV_MSG,code);
#endif
		}
	}
	spin_unlock(&bdev->lock);
}


/**
 * キーを取得し続ける高速スレッド
 */
static void gpio_poll(unsigned long data){
	struct input_polled_dev *pdev = (struct input_polled_dev*)data;
	struct input_device_server_dev *bdev = pdev->private;
	struct input_dev *idev = bdev->input;
	static volatile union key_down_uni key_current={0};

	spin_lock(&bdev->lock);

	//ここにHW上の状態が乗るように
	read_pin_values(&key_current);

	//ONになったものがあった場合に限り、キーをキューに追加
	set_key_override(&key_current,bdev->key_buffer);
	if(key_current.list.key_poweroff && bdev->gpio_timer_active){
		if(time_after(jiffies,bdev->watchdog_timer+msecs_to_jiffies(5*1000))){
			//5秒でウォッチドッグタイマー起動、強制的にOFF
			printk( KERN_ERR "%s : watchdog timer\n" ,DEV_MSG);
			//machine_halt();
			//machine_power_off();
			//machine_shutdown();
			write_pin(4,true);
			return 0;
		}
	}else{
		bdev->watchdog_timer = jiffies;
	}
	spin_unlock(&bdev->lock);

	mod_timer(&bdev->gpio_timer, jiffies + msecs_to_jiffies(POLL_INTERVAL_GPIO));
}

//fopen
static int _open(struct input_polled_dev *polled_dev) {
#ifdef DEBUG
	printk( KERN_INFO "%s : open\n" ,DEV_MSG);
#endif
	return 0;
}

//fclose
static int _close(struct input_polled_dev *polled_dev) {
#ifdef DEBUG
	printk( KERN_INFO "%s : release\n" ,DEV_MSG);
#endif
	return 0;
}

/*struct proc_dir_entry * auto_create_proc_entry(const char proc_path[]){
	struct proc_dir_entry *ent;
	char proc_dir_path[256],p[256];
	const char sep    = '/';
	const char final  = '\0';
	int i,j;
	struct file *fp;
	const int path_len = strlen(proc_path);
	strcpy(proc_dir_path,proc_path);
	for(i=0;i<path_len-1;i++){
		if (proc_dir_path[i] == sep) {
			proc_dir_path[i] = final;
			{
				sprintf(p, "/proc/%s", proc_dir_path );
#ifdef DEBUG
				printk( KERN_INFO "%s",p);
#endif
				fp = filp_open(p, O_RDONLY, 0);
				if(IS_ERR(fp)){
					printk( KERN_INFO "%s : create proc dir. %s\n" ,DEV_MSG ,proc_dir_path);
					ent = proc_mkdir(proc_dir_path, 0);
					if (!ent){
						printk( KERN_ERR "%s : cannot create proc dir error\n" ,DEV_MSG);
						return ent;
					}
				}else{
					filp_close(fp,NULL);
				}
			}
			proc_dir_path[i] = sep;
		}
	}
	return create_proc_entry(proc_path, 0, 0);;
}*/

static int _write_proc(char *st_b ,const int size ,struct semaphore *sem ,struct file *filp, const char *buf, unsigned long count){
	down( sem );
	const unsigned long copy_len = min((unsigned long)count,size);
	if ( copy_from_user( st_b, buf, copy_len ) ) {
		up( sem );
		printk( KERN_INFO "%s : copy_from_user failed\n", DEV_MSG);
		return -EFAULT;
	}
	if(0<size-copy_len){
		memset(&st_b[copy_len],0,size-copy_len);
	}
	up( sem );
	return count;//全部コピーできたことにしてしまう
}
static int _read_proc(const char *st_b ,struct semaphore *sem ,char *buf, char **start, off_t offset, int count, int *eof){
	down( sem );
	int data_length;
	int return_length;
	data_length = strlen(st_b);
	if ((offset + count) > data_length) {
		*eof = 1;
		if ((return_length = data_length - offset) <= 0) {
			up( sem );
			return 0; /* EOF後や、0バイトで呼ばれる場合がある */
		}
	}else {
		return_length = count;
	}
	strncpy(buf, st_b + offset, return_length);
	*start = buf + offset; /* 次のスタート位置を更新 */
	up( sem );
	return return_length;
}

static void _send_key(struct input_polled_dev *polled_dev,int code){
	struct input_device_server_dev *bdev = polled_dev->private;
	struct input_dev *idev = bdev->input;
	int value;
	value = 1;//key down
	input_report_key(idev,code,value);
	input_sync(idev);
	value = 0;//key up
	input_report_key(idev,code,value);
	input_sync(idev);
}

static struct semaphore g_proc_prog_sem;
static char g_progress_buf[ 16 ];
static int progress_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *data){
	const char *st_b = g_progress_buf; /* 転送元バッファをここに */
	static struct semaphore* sem = &g_proc_prog_sem;
	return _read_proc(st_b,sem,buf,start,offset,count,eof);
}
static int progress_write_proc(struct file *filp, const char *buf, unsigned long count, void *data){
	char *st_b = g_progress_buf; /* 転送先バッファをここに */
	const int size = sizeof(g_progress_buf);
	static struct semaphore* sem = &g_proc_prog_sem;
	int r = _write_proc(st_b,size,sem,filp,buf,count);
	_send_key((struct input_polled_dev*)data,UI_PROGRESS);
	return r;
}

static struct semaphore g_proc_error_sem;
static char g_error_buf[ 512 ];
static int detail_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *data){
	const char *st_b = g_error_buf; /* 転送元バッファをここに */
	static struct semaphore* sem = &g_proc_error_sem;
	return _read_proc(st_b,sem,buf,start,offset,count,eof);
}
static int detail_write_proc(struct file *filp, const char *buf, unsigned long count, void *data){
	char *st_b = g_error_buf; /* 転送先バッファをここに */
	const int size = sizeof(g_error_buf);
	static struct semaphore* sem = &g_proc_error_sem;
	return _write_proc(st_b,size,sem,filp,buf,count);
}

static struct semaphore g_proc_import_1_sem;
static char g_import_1_buf[ 64 ];
static int import_1_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *data){
	const char *st_b = g_import_1_buf; /* 転送元バッファをここに */
	static struct semaphore* sem = &g_proc_import_1_sem;
	return _read_proc(st_b,sem,buf,start,offset,count,eof);
}
static int import_1_write_proc(struct file *filp, const char *buf, unsigned long count, void *data){
	char *st_b = g_import_1_buf; /* 転送先バッファをここに */
	const int size = sizeof(g_import_1_buf);
	static struct semaphore* sem = &g_proc_import_1_sem;
	int r = _write_proc(st_b,size,sem,filp,buf,count);
	_send_key((struct input_polled_dev*)data,UI_IMPORT_1);
	return r;
}

static struct semaphore g_proc_import_2_sem;
static char g_import_2_buf[ 64 ];
static int import_2_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *data){
	const char *st_b = g_import_2_buf; /* 転送元バッファをここに */
	static struct semaphore* sem = &g_proc_import_2_sem;
	return _read_proc(st_b,sem,buf,start,offset,count,eof);
}
static int import_2_write_proc(struct file *filp, const char *buf, unsigned long count, void *data){
	char *st_b = g_import_2_buf; /* 転送先バッファをここに */
	const int size = sizeof(g_import_2_buf);
	static struct semaphore* sem = &g_proc_import_2_sem;
	int r = _write_proc(st_b,size,sem,filp,buf,count);
	_send_key((struct input_polled_dev*)data,UI_IMPORT_2);
	return r;
}

// USBODD処理用
static struct semaphore g_proc_usbodd_1_sem;
static char g_usbodd_1_buf[ 64 ];
static int usbodd_1_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *data){
	const char *st_b = g_usbodd_1_buf; /* 転送元バッファをここに */
	static struct semaphore* sem = &g_proc_usbodd_1_sem;
	return _read_proc(st_b,sem,buf,start,offset,count,eof);
}
static int usbodd_1_write_proc(struct file *filp, const char *buf, unsigned long count, void *data){
	char *st_b = g_usbodd_1_buf; /* 転送先バッファをここに */
	const int size = sizeof(g_usbodd_1_buf);
	static struct semaphore* sem = &g_proc_usbodd_1_sem;
	int r = _write_proc(st_b,size,sem,filp,buf,count);
	_send_key((struct input_polled_dev*)data,UI_USBODD_1);
	return r;
}

static struct semaphore g_proc_usbodd_2_sem;
static char g_usbodd_2_buf[ 64 ];
static int usbodd_2_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *data){
	const char *st_b = g_usbodd_2_buf; /* 転送元バッファをここに */
	static struct semaphore* sem = &g_proc_usbodd_2_sem;
	return _read_proc(st_b,sem,buf,start,offset,count,eof);
}
static int usbodd_2_write_proc(struct file *filp, const char *buf, unsigned long count, void *data){
	char *st_b = g_usbodd_2_buf; /* 転送先バッファをここに */
	const int size = sizeof(g_usbodd_2_buf);
	static struct semaphore* sem = &g_proc_usbodd_2_sem;
	int r = _write_proc(st_b,size,sem,filp,buf,count);
	_send_key((struct input_polled_dev*)data,UI_USBODD_2);
	return r;
}

static struct semaphore g_proc_expansion_sem;
static char g_expansion_buf[ 64 ];
static int expansion_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *data){
	const char *st_b = g_expansion_buf; /* 転送元バッファをここに */
	static struct semaphore* sem = &g_proc_expansion_sem;
	return _read_proc(st_b,sem,buf,start,offset,count,eof);
}
static int expansion_write_proc(struct file *filp, const char *buf, unsigned long count, void *data){
	char *st_b = g_expansion_buf; /* 転送先バッファをここに */
	const int size = sizeof(g_expansion_buf);
	static struct semaphore* sem = &g_proc_expansion_sem;
	int r = _write_proc(st_b,size,sem,filp,buf,count);
	_send_key((struct input_polled_dev*)data,UI_EXPANSION);
	return r;
}

static struct semaphore g_proc_watchdog_sem;
static int watchdog_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *data){
	struct input_polled_dev *pdev = (struct input_polled_dev*)data;
	struct input_device_server_dev *bdev = pdev->private;
	down( &g_proc_watchdog_sem );
	bdev->watchdog_timer = jiffies;//タイマー開始
	up( &g_proc_watchdog_sem );
	return 0;
}
static int watchdog_write_proc(struct file *filp, const char *buf, unsigned long count, void *data){
	struct input_polled_dev *pdev = (struct input_polled_dev*)data;
	struct input_device_server_dev *bdev = pdev->private;
	int power = simple_strtol( buf, NULL, 10 );
	down( &g_proc_watchdog_sem );
	bdev->watchdog_timer = jiffies;
	up( &g_proc_watchdog_sem );
	bdev->gpio_timer_active = (0 <= power);
	if(!bdev->gpio_timer_active){
		//watchdogの中止
		printk( KERN_ERR "%s : stop watchdog timer\n" ,DEV_MSG);
	}
	return count;
}

static struct semaphore g_proc_test_sem;
static char g_test_buf[ 16 ];
static int test_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *data){
	const char *st_b = g_test_buf; /* 転送元バッファをここに */
	static struct semaphore* sem = &g_proc_test_sem;
	return _read_proc(st_b,sem,buf,start,offset,count,eof);
}
static int test_write_proc(struct file *filp, const char *buf, unsigned long count, void *data){
	char *st_b = g_test_buf; /* 転送先バッファをここに */
	const int size = sizeof(g_test_buf);
	static struct semaphore* sem = &g_proc_test_sem;
	int r = _write_proc(st_b,size,sem,filp,buf,count);
	int code = simple_strtol(st_b,NULL,10);
	if(0<code){
		_send_key((struct input_polled_dev*)data,code);
	}
	return r;
}

static struct semaphore g_proc_dmr_sem;
static char g_dmr_buf[ 64 ];
static int dmr_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *data){
	const char *st_b = g_dmr_buf; /* 転送元バッファをここに */
	static struct semaphore* sem = &g_proc_dmr_sem;
	return _read_proc(st_b,sem,buf,start,offset,count,eof);
}
static int dmr_write_proc(struct file *filp, const char *buf, unsigned long count, void *data){
	char *st_b = g_dmr_buf; /* 転送先バッファをここに */
	const int size = sizeof(g_dmr_buf);
	static struct semaphore* sem = &g_proc_dmr_sem;
	int r = _write_proc(st_b,size,sem,filp,buf,count);
	_send_key((struct input_polled_dev*)data,UI_DMR);
	return r;
}

static struct semaphore g_proc_notify_sem;
static char g_notify_buf[ 64 ];
static int notify_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *data){
	const char *st_b = g_notify_buf; /* 転送元バッファをここに */
	static struct semaphore* sem = &g_proc_notify_sem;
	return _read_proc(st_b,sem,buf,start,offset,count,eof);
}
static int notify_write_proc(struct file *filp, const char *buf, unsigned long count, void *data){
	char *st_b = g_notify_buf; /* 転送先バッファをここに */
	const int size = sizeof(g_notify_buf);
	static struct semaphore* sem = &g_proc_notify_sem;
	int r = _write_proc(st_b,size,sem,filp,buf,count);
	_send_key((struct input_polled_dev*)data,UI_NOTIFY);
	return r;
}

struct input_polled_dev *g_pdev;

static void __exit _exit() {
	struct input_polled_dev *pdev = g_pdev;
	struct input_device_server_dev *bdev = pdev->private;
	struct input_dev *idev = bdev->input;
	bdev->gpio_timer_active = false;
	del_timer_sync(&bdev->gpio_timer);
	input_unregister_polled_device(pdev);
	kfree(bdev);

	on_key_cleanup();
#ifdef DEBUG
	printk( KERN_INFO "%s : _exit\n" ,DEV_MSG);
#endif
}

static int __init _init() {
	struct input_polled_dev *pdev;
	struct input_dev *idev;
	struct input_device_server_dev *bdev;
	int result = -EINVAL,i;

#ifdef DEBUG
	printk( KERN_INFO "%s : _init\n" ,DEV_MSG);
#endif

	bdev = kzalloc(sizeof(struct input_device_server_dev),GFP_KERNEL);
	if (!bdev) {
		printk("%s : no memory for private data\n",DEV_MSG);
		result = -ENOMEM;
		goto err;
	}
	memset(bdev,0,sizeof(struct input_device_server_dev));

	pdev = input_allocate_polled_device();
	if (!pdev) {
		printk("%s : cannot allocate polled device\n",DEV_MSG);
		result = -ENOMEM;
		goto err;
	}

	pdev->private = bdev;
	pdev->open = _open;
	pdev->close = _close;
	pdev->poll  = _poll;
	pdev->poll_interval     = POLL_INTERVAL_MIN;
	pdev->poll_interval_min = POLL_INTERVAL_MIN;
	pdev->poll_interval_max = POLL_INTERVAL_MAX;

	idev = pdev->input;

	idev->name = DEV_NAME;
	idev->phys = "A/Fake/Path";
	idev->id = info;
	idev->rep[REP_DELAY]  = POLL_INTERVAL_MAX;
	idev->rep[REP_PERIOD] = POLL_INTERVAL_MIN;
	set_bit(EV_KEY, idev->evbit);
	set_bit(EV_REP, idev->evbit);
	for(i=0;i<key_down_all_list_count;i++){
		set_bit(key_code_array[i], idev->keybit);
	}

	bdev->input = idev;
	bdev->watchdog_timer = 0;
	bdev->last_keydown = 0;
	bdev->key_buffer = kzalloc(sizeof(union key_down_uni),GFP_KERNEL);//kzallocはゼロ埋めも行う
	spin_lock_init( &bdev->lock );

	on_key_init();

	{
		//////////////////////////
		//procディレクトリの準備
		//初回起動時はauto_create_proc_entryがうまくいかない
		struct proc_dir_entry *ent = proc_mkdir(PROC_DIR_PATH, 0);
		if (!ent){
			printk( KERN_ERR "%s : cannot create proc dir error\n" ,DEV_MSG);
			result = -EBUSY;
			goto err;
		}
	}
	{
		//////////////////////////
		//Progress
		sema_init( &g_proc_prog_sem, 1 );
		//struct proc_dir_entry *ent = auto_create_proc_entry(PROC_PROGRESS_PATH);
		struct proc_dir_entry *ent = create_proc_entry(PROC_PROGRESS_PATH,0,0);
		if (!ent){
			printk( KERN_ERR "%s : cannot create proc error\n" ,DEV_MSG);
			result = -EBUSY;
			goto err;
		}
		ent->read_proc  = progress_read_proc;
		ent->write_proc = progress_write_proc;
		ent->data = pdev;
	}
	{
		//////////////////////////
		//Detail
		sema_init( &g_proc_error_sem, 1 );
		//struct proc_dir_entry *ent = auto_create_proc_entry(PROC_DETAIL_PATH);
		struct proc_dir_entry *ent = create_proc_entry(PROC_DETAIL_PATH,0,0);
		if (!ent){
			printk( KERN_ERR "%s : cannot create proc error\n" ,DEV_MSG);
			result = -EBUSY;
			goto err;
		}
		ent->read_proc  = detail_read_proc;
		ent->write_proc = detail_write_proc;
		ent->data = pdev;
	}
	{
		//////////////////////////
		//Import1
		sema_init( &g_proc_import_1_sem, 1 );
		//struct proc_dir_entry *ent = auto_create_proc_entry(PROC_IMPORT_1_PATH);
		struct proc_dir_entry *ent = create_proc_entry(PROC_IMPORT_1_PATH,0,0);
		if (!ent){
			printk( KERN_ERR "%s : cannot create proc error\n" ,DEV_MSG);
			result = -EBUSY;
			goto err;
		}
		ent->read_proc  = import_1_read_proc;
		ent->write_proc = import_1_write_proc;
		ent->data = pdev;
	}
	{
		//////////////////////////
		//Import2
		sema_init( &g_proc_import_2_sem, 1 );
		//struct proc_dir_entry *ent = auto_create_proc_entry(PROC_IMPORT_2_PATH);
		struct proc_dir_entry *ent = create_proc_entry(PROC_IMPORT_2_PATH,0,0);
		if (!ent){
			printk( KERN_ERR "%s : cannot create proc error\n" ,DEV_MSG);
			result = -EBUSY;
			goto err;
		}
		ent->read_proc  = import_2_read_proc;
		ent->write_proc = import_2_write_proc;
		ent->data = pdev;
	}

	{
		//////////////////////////
		//usbodd1
		sema_init( &g_proc_usbodd_1_sem, 1 );
		//struct proc_dir_entry *ent = auto_create_proc_entry(PROC_USBODD_1_PATH);
		struct proc_dir_entry *ent = create_proc_entry(PROC_USBODD_1_PATH,0,0);
		if (!ent){
			printk( KERN_ERR "%s : cannot create proc error\n" ,DEV_MSG);
			result = -EBUSY;
			goto err;
		}
		ent->read_proc  = usbodd_1_read_proc;
		ent->write_proc = usbodd_1_write_proc;
		ent->data = pdev;
	}
	{
		//////////////////////////
		//usbodd2
		sema_init( &g_proc_usbodd_2_sem, 1 );
		//struct proc_dir_entry *ent = auto_create_proc_entry(PROC_USBODD_2_PATH);
		struct proc_dir_entry *ent = create_proc_entry(PROC_USBODD_2_PATH,0,0);
		if (!ent){
			printk( KERN_ERR "%s : cannot create proc error\n" ,DEV_MSG);
			result = -EBUSY;
			goto err;
		}
		ent->read_proc  = usbodd_2_read_proc;
		ent->write_proc = usbodd_2_write_proc;
		ent->data = pdev;
	}

	{
		//////////////////////////
		//Expansion
		sema_init( &g_proc_expansion_sem, 1 );
		//struct proc_dir_entry *ent = auto_create_proc_entry(PROC_EXPANSION_PATH);
		struct proc_dir_entry *ent = create_proc_entry(PROC_EXPANSION_PATH,0,0);
		if (!ent){
			printk( KERN_ERR "%s : cannot create proc error\n" ,DEV_MSG);
			result = -EBUSY;
			goto err;
		}
		ent->read_proc  = expansion_read_proc;
		ent->write_proc = expansion_write_proc;
		ent->data = pdev;
	}
	{
		//////////////////////////
		//Expansion2
		sema_init( &g_proc_expansion_sem, 1 );

		//struct proc_dir_entry *ent = auto_create_proc_entry(PROC_EXPANSION_2_PATH);
		struct proc_dir_entry *ent = create_proc_entry(PROC_EXPANSION_2_PATH,0,0);
		if (!ent){
			printk( KERN_ERR "%s : cannot create proc error\n" ,DEV_MSG);
			result = -EBUSY;
			goto err;
		}
		ent->read_proc  = expansion_read_proc;
		ent->write_proc = expansion_write_proc;
		ent->data = pdev;
	}
	{
		//////////////////////////
		//Watchdog
		sema_init( &g_proc_watchdog_sem, 1 );
		//struct proc_dir_entry *ent = auto_create_proc_entry(PROC_WATCHDOG_PATH);
		struct proc_dir_entry *ent = create_proc_entry(PROC_WATCHDOG_PATH,0,0);
		if (!ent){
			printk( KERN_ERR "%s : cannot create proc error\n" ,DEV_MSG);
			result = -EBUSY;
			goto err;
		}
		ent->read_proc  = watchdog_read_proc;
		ent->write_proc = watchdog_write_proc;
		ent->data = pdev;
	}
	{
		//////////////////////////
		//Test
		sema_init( &g_proc_test_sem, 1 );
		//struct proc_dir_entry *ent = auto_create_proc_entry(PROC_TEST_PATH);
		struct proc_dir_entry *ent = create_proc_entry(PROC_TEST_PATH,0,0);
		if (!ent){
			printk( KERN_ERR "%s : cannot create proc error\n" ,DEV_MSG);
			result = -EBUSY;
			goto err;
		}
		ent->read_proc  = test_read_proc;
		ent->write_proc = test_write_proc;
		ent->data = pdev;
	}
	{
		//////////////////////////
		//DMR
		sema_init( &g_proc_dmr_sem, 1 );
		//struct proc_dir_entry *ent = auto_create_proc_entry(PROC_DMR_STATE_PATH);
		struct proc_dir_entry *ent = create_proc_entry(PROC_DMR_STATE_PATH,0,0);
		if (!ent){
			printk( KERN_ERR "%s : cannot create proc error\n" ,DEV_MSG);
			result = -EBUSY;
			goto err;
		}
		ent->read_proc  = dmr_read_proc;
		ent->write_proc = dmr_write_proc;
		ent->data = pdev;
	}
	{
		//////////////////////////
		//NOTIFY
		sema_init( &g_proc_notify_sem, 1 );
		//struct proc_dir_entry *ent = auto_create_proc_entry(PROC_NOTIFY_STATE_PATH);
		struct proc_dir_entry *ent = create_proc_entry(PROC_NOTIFY_STATE_PATH,0,0);
		if (!ent){
			printk( KERN_ERR "%s : cannot create proc error\n" ,DEV_MSG);
			result = -EBUSY;
			goto err;
		}
		ent->read_proc  = notify_read_proc;
		ent->write_proc = notify_write_proc;
		ent->data = pdev;
	}

	//スレッド生成
	/*bdev->key_thread = kthread_run(get_key_thread, (void *)pdev, "tzkeywatcher");
	if(!bdev->key_thread){
		result = -EBUSY;
		goto err;
	}*/

	init_timer(&bdev->gpio_timer);
	bdev->gpio_timer.function = gpio_poll;
	bdev->gpio_timer.data = (long) pdev;
	bdev->gpio_timer.expires = jiffies + msecs_to_jiffies(POLL_INTERVAL_GPIO);
	add_timer(&bdev->gpio_timer);

	input_register_polled_device(pdev);
	g_pdev = pdev;
	printk("%s %s Ver.%s installed.\n", DRIVER_NAME, AUTHOR, BUFFALO_DRIVER_VER);
	return 0;

err:
	//エラー処理系
	_exit();//ロールバック
	return result;
}

late_initcall(_init);
module_exit(_exit);

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DRIVER_NAME);
MODULE_LICENSE("GPL");
MODULE_VERSION(BUFFALO_DRIVER_VER);
